# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import sys
from datetime import date

# The full version, including alpha/beta/rc tags
from importlib.metadata import version as get_version
from pathlib import Path

root_path = str(Path(__file__).resolve().parent.parent)
sys.path.insert(0, root_path)
sys.setrecursionlimit(1500)

# -- Project information -----------------------------------------------------

project = "hydra-config"
copyright = f"{date.today().year}, Aaron Young"
author = "Aaron Young"
release = get_version(project)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "autoapi.extension",
    "myst_parser",
    "sphinx_copybutton",
    "sphinxcontrib.video",
    "sphinxcontrib.googleanalytics",
]

# autoapi config
autoapi_type = "python"
autoapi_dirs = ["../hydra_config"]
autoapi_options = [
    "members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
    # "inherited-members"  # Doesn't work with viewcode extension
]
autoapi_ignore = []
autoapi_keep_files = False
autoapi_keep_files = True
autoapi_generate_api_docs = True
autoapi_add_toctree_entry = False
autoapi_root = "reference/api/"
# autoapi_template_dir = "_templates"
# autoapi_member_order = "groupwise"

# suppress_warnings = ["autoapi"]

add_module_names = False
autodoc_typehints = "none"

# Napoleon
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_use_ivar = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# MyST
myst_heading_anchors = 7

viewcode_enable_epub = True

# Display todos by setting to True
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Google Analytics
googleanalytics_id = "G-4SJ32LRS8B"


# -- Options for HTML output -------------------------------------------------

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_favicon = "_static/doc.png"
html_theme_options = {
    "announcement": """
        <a style=\"text-decoration: none; color: white;\"
           href=\"https://AaronYoung5.github.io\">
           Check out my other projects!
        </a>
    """,
    "sidebar_hide_name": True,
    "light_logo": "doc.png",
    "dark_logo": "doc.png",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "css/custom.css",
]



"""This module contains custom resolvers for OmegaConf."""

import re
from pathlib import Path
from typing import Any, Dict, Optional

from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.errors import ConfigKeyError

from hydra_config.config import HydraContainerConfig
from hydra_config.utils import safe_eval

# =============================================================================


def register_new_resolver(name: str, replace: bool = True, **kwargs):
    """Register a new resolver with OmegaConf.

    Args:
        name (str): The name of the resolver.
        replace (bool): Whether to replace the resolver if it already exists.
        **kwargs: Additional keyword arguments to pass to
            ``OmegaConf.register_new_resolver``.
    """

    def decorator(fn):
        OmegaConf.register_new_resolver(name, fn, replace=replace, **kwargs)
        return fn

    return decorator


@register_new_resolver("search")
def search_resolver(
    key: str | None = None,
    /,
    mode: Optional[str] = "value",
    *,
    depth: int = 0,
    _parent_: DictConfig,
) -> Any:
    """This method will recursively search up the parent chain for the key and return
    the value. If the key is not found, will raise a KeyError.

    For instance, a heavily nested value might want to access a value some level
    higher but it may be hazardous to use relative paths (i.e. ${..key}) since
    the config may be changed. Instead, we'll search up for a specific key to set the
    value to. Helpful for setting unique names for an object in a nested config.

    Note:
        This technically uses hidden attributes (i.e. _parent).

    Args:
        key (str | None): The key to search for. Could be none (like when mode is
            "parent_key").
        mode (Optional[str]): The mode to use. Defaults to "value". Available modes:
            - "value": Will return the value of the found key. Key must be set.
            - "parent_key": Will return the parent's key. If key is None, won't do
            any recursion and will return the parent's key.
            - "path": Will return the path to the key.
        depth (Optional[int]): The depth of the search. Used internally
            in this method and unsettable from the config. Avoids checking the parent
            key.
        _parent_ (DictConfig): The parent config to search in.
    """
    if _parent_ is None:
        # Parent will be None if we're at the top level
        raise ConfigKeyError(f"Key {key} not found in parent chain.")

    if mode == "value":
        if key in _parent_:
            # If the key is in the parent, we'll return the value
            return _parent_[key]
        else:
            # Otherwise, we'll keep searching up the parent chain
            return search_resolver(
                key, mode=mode, depth=depth + 1, _parent_=_parent_._parent
            )
    elif mode == "parent_key":
        if key is None:
            # If the key is None, we'll return the parent's key
            assert _parent_._key() is not None, "Parent key is None."
            return _parent_._key()
        elif _parent_._key() == key:
            assert _parent_._parent._key() is not None, "Parent key is None."
            return _parent_._parent._key()

        if depth != 0 and isinstance(_parent_, DictConfig) and key in _parent_:
            # If we're at a key that's not the parent and the parent has the key we're
            # looking for, we'll return the parent
            return search_resolver(None, mode=mode, depth=depth + 1, _parent_=_parent_)
        else:
            # Otherwise, we'll keep searching up the parent chain
            return search_resolver(
                key, mode=mode, depth=depth + 1, _parent_=_parent_._parent
            )
    elif mode == "path":
        if key in _parent_:
            # If the key is in the parent, we'll return the path
            return _parent_._get_full_key(key)
        else:
            # Otherwise, we'll keep searching up the parent chain
            return search_resolver(
                key, mode=mode, depth=depth + 1, _parent_=_parent_._parent
            )


@register_new_resolver("parent")
def parent_resolver(
    key: Optional[str] = None, mode: str = "parent_key", *, _parent_: DictConfig
) -> Any:
    """This resolver is a wrapper around the search resolver with a default mode of
    "parent_key". This will return the parent's key."""
    return search_resolver(key, mode=mode, _parent_=_parent_)


@register_new_resolver("eval")
def eval_resolver(key: str, /, *, _root_: DictConfig) -> Any:
    """This resolver will evaluate the key as a python expression. This is useful for
    evaluating `any` expression. The resolver calls
    :meth:`~hydra_config.utils.safe_eval`, so only a (safe) subset of python is
    available."""

    try:
        return safe_eval(key)
    except Exception as e:
        _root_._format_and_raise(
            key=key,
            value=key,
            msg=f"Error evaluating expression '{key}': {e}",
            cause=e,
        )


@register_new_resolver("glob")
def glob_resolver(
    pattern: str,
    config: Optional[DictConfig | ListConfig | str] = None,
    /,
    *,
    _root_: DictConfig,
) -> ListConfig | DictConfig:
    """This resolver will return a list of keys that match the pattern. This is useful
    for selecting a subset of keys in a config. The pattern should be a regex pattern.
    If the config is a string, it will be treated as a dotpath to a config."""

    if config is None:
        config = _root_

    if isinstance(config, str):
        config = OmegaConf.select(_root_, config)
    if isinstance(config, DictConfig):
        return {k: v for k, v in config.items() if re.match(pattern, k)}
    if isinstance(config, ListConfig):
        return [v for v in config if re.match(pattern, v)]


@register_new_resolver("hydra_select")
def hydra_select(
    key: str, default: Optional[Any] = None, /, *, _root_: DictConfig
) -> Any | None:
    """This is similar to the regular hydra resolver, but this won't through an error
    if the global hydra config is unset. Instead, it will return another interpolation
    using dotpath notation directly. As in, ${hydra_select:runtime.choices.test}, if
    HydraConfig is unset, will return ${hydra.runtime.choices.test}."""
    from hydra.core.hydra_config import HydraConfig

    try:
        return OmegaConf.select(HydraConfig.get(), key, default=default)
    except ValueError:
        return OmegaConf.select(
            _root_, f"hydra.{key}", default=default, throw_on_missing=True
        )


@register_new_resolver("path")
def path_resolver(*parts: str) -> Path:
    """Simple resolver to join paths together."""
    return Path(*parts)


@register_new_resolver("read")
def read_resolver(path: str) -> str:
    """Simple resolver to read a file."""
    with open(path, "r") as f:
        return f.read()


@register_new_resolver("custom")
def custom_resolver(target: str, default: Optional[Any] = None, /):
    """This resolver is a wrapper around the select resolver to grab an entry in a
    parent :attr:`~hydra_config.config.HydraContainerConfig.custom` field.
    """
    return f"${{oc.select:${{search:custom,'path'}}.{target}, {default}}}"


@register_new_resolver("float_to_str")
def float_to_str_resolver(value: float) -> str:
    """This resolver will convert a float to a string. This is useful for saving
    floats as strings in a config for a filename. All periods will be replaced with
    'p' and all negatives will be replaced with 'n'; e.g. ``-0.5`` will be converted
    to ``'n0p5'``."""
    return str(value).replace(".", "p").replace("-", "n")


@register_new_resolver("locate")
def locate_resolver(fn: str, /, cast_to: Optional[str] = None) -> Any:
    """This resolver will locate an object using the :meth:`hydra.utils.get_object`
    method."""
    from hydra.utils import get_object

    obj = get_object(fn)
    if cast_to is not None:
        if cast_to == "str":
            return str(obj)
        elif cast_to == "int":
            return int(obj)
        elif cast_to == "float":
            return float(obj)
        elif cast_to == "bool":
            return bool(obj)
        else:
            raise ValueError(f"Invalid cast_to value: {cast_to}")
    return obj


@register_new_resolver("getitem")
def getitem_resolver(obj: ListConfig, key: int) -> Any:
    """This resolver will get an item from a list."""
    return obj[key]


@register_new_resolver("target")
def target_resolver(target: str, /, *args) -> Dict[str, Any]:
    """This is a resolver which serves as a proxy for the _target_ attribute used
    in hydra. Basically :attr:`target` will be defined as ``_target_`` and the rest of
    the attributes will be passed as arguments to the target. You should always
    default to using ``_target_`` directly in your config, but because interpolations
    `may` be resolved prior to or instead of instantiate, it may be desired to resolve
    interpolations before instantiations."""
    return {"_target_": target, "_args_": args}


@register_new_resolver("instantiate")
def instantiate_resolver(target: str | DictConfig, /, *args, _root_: DictConfig) -> Any:
    """This resolver will instantiate a target using the HydraContainerConfig. If the
    target is a string, it will be resolved using the target_resolver. If the target is
    a DictConfig, it will be passed directly to the HydraContainerConfig.instantiate
    method."""
    try:
        if isinstance(target, str):
            target = target_resolver(target, *args)
        return HydraContainerConfig.instantiate(target)
    except Exception as e:
        _root_._format_and_raise(
            key=target,
            value=target,
            msg=f"Error instantiating target '{target}': {e}",
            cause=e,
        )



"""``hydra_config`` is a package that provides utilities to simplify the usage of Hydra
in your project."""

from hydra.core.utils import setup_globals

from hydra_config.config import HydraContainerConfig, config_wrapper
from hydra_config.resolvers import register_new_resolver
from hydra_config.utils import HydraFlagWrapperMeta, run_hydra

# Call explicitly setup_globals() to ensure that Hydra is properly initialized
setup_globals()

__all__ = [
    "config_wrapper",
    "HydraContainerConfig",
    "run_hydra",
    "HydraFlagWrapperMeta",
    "register_new_resolver",
]



"""This module contains utility functions for use with Hydra."""

import argparse
import enum
import re
from functools import partial
from pathlib import Path
from typing import Any, Callable, Concatenate, Dict, List, Optional

import hydra
import RestrictedPython
import RestrictedPython.Eval
from hydra.utils import get_object
from omegaconf import DictConfig, OmegaConf

from hydra_config.config import HydraContainerConfig

# =============================================================================


def run_hydra(
    main_fn: Optional[
        Callable[[Concatenate[HydraContainerConfig, ...]], None]
    ] = lambda *_, **__: None,
    /,
    *,
    parser: argparse.ArgumentParser = argparse.ArgumentParser(),
    config_path: Path | str = Path.cwd() / "configs",
    config_name: str = "base",
    instantiate: bool = True,
    **kwargs,
):
    """This function is the main entry point for the hydra application.

    The benefits of using this setup rather than the compose API is that we can
    use the sweeper and launcher APIs, which are not available in the compose API.

    An additional ``--hydra-help`` flag is added to the parser to print the hydra help
    message when passed.

    Args:
        main_fn (Callable[[Concatenate[[HydraContainerConfig], ...], None]): The
            main function to be called after the hydra configuration is parsed. It
            should take the config as an argument and kwargs which correspond to the
            argument parser returns. We don't return the config directly because hydra
            allows multi-run sweeps and it doesn't make sense to return multiple
            configs in this case.

            Example:

            .. code-block:: python

                def main(config: HydraContainerConfig, *, verbose: int):
                    pass

                parser = argparse.ArgumentParser()
                parser.add_argument("--verbose", type=int, default=0)

                run_hydra(main_fn=main, parser=parser)

    Keyword Args:
        parser (argparse.ArgumentParser): The parser to use for the hydra
            application. If None, a new parser will be created.
        config_path (Path | str): The path to the config directory. This should be the
            absolute path to the directory containing the config files. By default,
            this is set to the current working directory.
        config_name (str): The name of the config file to use. This should be the
            name of the file without the extension. By default, this is set to
            "base".
        instantiate (bool): Whether to instantiate the config. If False, create
            will be used.
        kwargs: Additional keyword arguments to pass to the instantiate function.
    """

    # Add one default argument for the --hydra-help message
    parser.add_argument(
        "--no-instantiate",
        action="store_false",
        dest="instantiate",
        help="Don't instantiate the config.",
    )
    parser.add_argument(
        "--hydra-help", action="store_true", help="Print the hydra help message."
    )

    def hydra_argparse_override(fn: Callable, /):
        """This function allows us to add custom argparse parameters prior to hydra
        parsing the config.

        We want to set some defaults for the hydra config here. This is a workaround
        in a way such that we don't

        Note:
            Augmented from hydra discussion #2598.
        """
        import sys
        from functools import partial

        parsed_args, unparsed_args = parser.parse_known_args()

        # Move --hydra-help to unparsed_args if it's present
        # Hydra has a weird bug (I think) that doesn't allow overrides when
        # --hydra-help is passed, so remove all unparsed arguments if --hydra-help
        # is passed.
        if parsed_args.hydra_help:
            unparsed_args = ["--hydra-help"]
        del parsed_args.hydra_help

        # By default, argparse uses sys.argv[1:] to search for arguments, so update
        # sys.argv[1:] with the unparsed arguments for hydra to parse (which uses
        # argparse).
        sys.argv[1:] = unparsed_args

        return partial(fn, **vars(parsed_args))

    @hydra.main(
        config_path=str(config_path), config_name=config_name, version_base=None
    )
    @hydra_argparse_override
    def main(config: DictConfig, instantiate: bool = instantiate, **fn_kwargs):
        if instantiate:
            config = HydraContainerConfig.instantiate(config, **kwargs)

        return main_fn(config, **fn_kwargs)

    main()


# =============================================================================


def safe_eval(src: str, additional_vars: Dict[str, Any] = {}) -> Any:
    """
    Evaluate a string containing a Python expression in a safe manner.

    This function uses
    `RestrictedPython <https://restrictedpython.readthedocs.io/en/latest/>`_ to evaluate
    the expression, only allowing certain built-in functions and types, and any
    additional variables provided. It prevents execution of arbitrary code or access to
    unauthorized functions and methods.

    A number of built-in functions are supported, as provided by ``utility_builtins``,
    ``safe_builtins``, and ``limited_builtins`` from
    `here <https://restrictedpython.readthedocs.io/en/latest/usage/\
        api.html#restricted-builtins>`_.

    Args:
        src (str): The source code to evaluate.
        additional_vars (Dict[str, Any]): A dictionary of additional variables or
            functions to include in the evaluation environment.

            Warning:
                This can be unsafe if the variables are not properly sanitized, thus
                ``additional_vars`` should be used with caution.

    Returns:
        Any: The result of the evaluated expression.

    Example:
        >>> safe_eval("1 + 2")
        3
        >>> safe_eval("max([1, 2, 3])")
        3
        >>> safe_eval("math.sqrt(a)", {'a': 16})
        4.0
    """
    safe_globals = {
        **RestrictedPython.utility_builtins,
        **RestrictedPython.safe_builtins,
        **RestrictedPython.limited_builtins,
        "_getiter_": RestrictedPython.Eval.default_guarded_getiter,
    }
    byte_code = RestrictedPython.compile_restricted(
        src, filename="<inline code>", mode="eval"
    )
    return eval(byte_code, safe_globals, additional_vars)


def glob(key: str, flattened: bool, _root_: DictConfig) -> Dict:
    """This resolver will glob a key in the config. This is useful for finding all keys
    that match a pattern. This is particularly useful for finding all keys that match a
    pattern in a nested config. This is effectively select, but allows ``*`` to be used
    as a wildcard.

    This method works by finding all ``*`` in the key and then iterating over all
    subsequent keys that match the globbed pattern.

    Note:
        yaml files aren't necessarily built to support globbing (like xml), so
        this method is fairly slow and should be used sparingly.

    Note:
        List indexing is limited in support. To index an element in a list, you
        must use bracket notation. For instance, ``a[0].b`` is supported, but ``a.0.b``
        is not.

    Args:
        key (str): The key to glob. This is a dotlist key, like ``a.b.*``. Multiple
            globs can be used, like ``a.*.c.*.d.*``. Globs in keys can be used, as
            well, such as ``a.ab*.c``
        flatten (bool): If true, the output will be a dict of the leaf keys and
            the accumulated values if there are like leaf keys. If False, the
            output will be a nested dict. Defaults to False.
        _root_ (DictConfig): The root config.
    """

    def recursive_glob(config: DictConfig | Any, keys: List[str]) -> Dict:
        if not keys or not isinstance(config, DictConfig):
            return config

        # We'll loop over all the keys and find each match with the passed key/pattern
        result = {}
        current_key = keys[0].replace("*", ".*")
        for sub_key, sub_value in config.items():
            if sub_value is None:  # Skip None values, probably optionals
                continue

            if match := re.fullmatch(current_key, sub_key):
                # If it's a match, we'll recursively glob the next key
                matched_key = match.group()
                result[matched_key] = recursive_glob(sub_value, keys[1:])

        # This adds support for direct indexing. This is currently the only supported
        # way to do list accessing for globbing. To check, we'll clean the parentheses
        # and see if the key exists in the config as is.
        # NOTE: this is done after the recursive globbing in case the the key is found
        # earlier
        for cleaned_key in re.sub(r"^\((.*)\)$", r"\1", current_key).split("|"):
            if cleaned_key in result:
                continue

            if sub_value := OmegaConf.select(config, cleaned_key):
                # remove the brackets from the key
                cleaned_key = re.sub(r"^\((.*)\)$", r"\1", cleaned_key)
                result[cleaned_key] = recursive_glob(sub_value, keys[1:])

        return result

    def flatten(
        data: Dict[str, Any], values: Dict[str, List[Any]] = {}
    ) -> Dict[str, Any]:
        """This will flatten the nested dict to a flat dict where each key is a leaf
        key of the nested dict and the value is a list of all the values that were
        accumulated to that leaf key."""
        for k, v in data.items():
            if isinstance(v, dict):
                flatten(v, values)
            else:
                values.setdefault(k, [])
                values[k].append(v)
        return values

    # Glob the key(s)
    globbed = recursive_glob(_root_, key.split("."))

    # Return the flattened or nested dict
    return flatten(globbed) if flattened else globbed


def merge_with_kwargs(
    config: DictConfig,
    *,
    instantiate: bool = True,
    **kwargs,
) -> DictConfig:
    """This method will merge the kwargs into the config. This is useful for merging
    "late", as in after the config has been resolved (not instantiated). By specifying
    the merge to happen at instantiation time rather than at resolution time, it gives
    more freedom in defining overrides within the config.

    This is intended to be called from a yaml config file like:

    .. code-block:: yaml

        config_to_merge_late:
            _target_: <path_to>.merge_with_kwargs
            _recursive_: False
            config: ${...} # this is what the kwargs are merged into
            kwarg1: value1
            kwarg2: value2
            ...

    Note:
        You may want ``_recursive_=False`` (as above) to avoid instantiating the config
        before merging the kwargs. If you want to override a config attribute in the
        config object which is instantiated (i.e. is a partial), you won't have access
        to the config attribute (only the partial object), so you would want
        ``_recursive_=False``. Simpler cases can just use ``_recursive_=True``.

    Args:
        config (DictConfig): The config to merge the kwargs into.

    Keyword Args:
        kwargs: The kwargs to merge into the config.
    """
    config = OmegaConf.unsafe_merge(config, kwargs)

    if instantiate:
        return HydraContainerConfig.instantiate(config)
    return config


# =============================================================================


def instance_wrapper(
    *,
    instance: Any,
    key: Optional[str] = None,
    locate: bool = False,
    eval: bool = False,
    setitem: bool = False,
    **kwargs,
):
    """Wraps a class instance to allow setting class attributes after initialization.

    This utility is useful when not all attributes are available during class
    instantiation, allowing attributes to be set post-construction using either
    direct assignment, item setting, or attribute modification based on optional flags.

    Args:
        instance (Any): The class instance to wrap.
        key (Optional[str], optional): If provided, fetches the specified attribute
            from the instance to modify. Defaults to None.
        locate (bool, optional): If True, attempts to resolve attribute names
            dynamically (e.g., via object lookup). Defaults to False.
        eval (bool, optional): If True, evaluates attribute values using safe_eval
            before assignment. Defaults to False.
        setitem (bool, optional): If True, uses item assignment
            (e.g., ``instance[key]``) instead of ``setattr``. Defaults to False.
        **kwargs: Key-value pairs of attributes to set on the instance.

    Returns:
        Any: The modified instance.

    Raises:
        ValueError: If there is an error while setting an attribute.

    Example:

        .. code-block:: yaml

            obj_to_instantiate:
                _target_: <path_to>.instance_wrapper
                instance:
                    _target_: <class>
                    _args_: [arg1, arg2]
                    init_arg1: value1
                    init_arg2: value2
                set_arg1: value1
                set_arg2: value2

        For partial instantiation:

        .. code-block:: yaml

            partial_obj_to_instantiate:
                _target_: <path_to>.instance_wrapper
                instance:
                    _target_: <class>
                    _partial_: True
                    _args_: [arg1, arg2]
                    init_arg3: '???' # Set later
                set_arg1: value1
                set_arg2: value2

    """

    def setattrs(instance, **kwargs):
        try:
            for key, value in kwargs.items():
                if callable(value):
                    value = value()

                if locate:
                    key = get_object(key)
                if eval:
                    key = safe_eval(key)

                if isinstance(value, dict):
                    setattrs(getattr(instance, key), **value)
                elif setitem:
                    instance[key] = value
                else:
                    setattr(instance, key, value)
        except Exception as e:
            raise ValueError(f"Error when setting attribute {key=} to {value=}: {e}")
        return instance

    if isinstance(instance, partial):
        # If the instance is a partial, we'll setup a wrapper such that once the
        # partial is actually instantiated, we'll set the attributes of the instance
        # with the kwargs.
        partial_instance = instance
        config_kwargs = kwargs

        def wrapper(*args, **kwargs):
            if key is not None:
                instance = getattr(partial_instance, key)

            instance = partial_instance(*args, **kwargs)
            return setattrs(instance, **config_kwargs)

        return wrapper
    else:
        if key is not None:
            instance = getattr(instance, key)
        return setattrs(instance, **kwargs)


class HydraFlagWrapperMeta(enum.EnumMeta):
    """This is a simple metaclass to allow for the use of the | operator to combine
    flags. This means you can simply put ``flag1 | flag2`` in the yaml file and it will
    be combined into a single flag.

    The following forms are supported and any combination thereof:

    - ``flag1 | flag2 | flag3 | ...``
    - ``flag1|flag2|flag3|...``
    - ``flag1``
    """

    def __getitem__(cls, item):
        """This method allows for the use of the | operator to combine flags. This
        means you can simply put ``flag1 | flag2`` in the yaml file and it will be
        combined into a single flag.
        """

        if isinstance(item, str) and "|" in item:
            from functools import reduce

            items = [cls.__getitem__(i.strip()) for i in item.split("|")]
            return reduce(lambda x, y: x | y, items)
        return super().__getitem__(item)



"""This module provides a base class for working with Hydra configs."""

import enum
import types
from copy import deepcopy
from dataclasses import dataclass, field, fields, make_dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Self, Tuple

import hydra_zen as zen
import yaml
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, DictConfig, ListConfig, OmegaConf

# =============================================================================


def config_wrapper(cls=None, /, **kwargs):
    """This is a wrapper of the dataclass decorator that adds the class to the hydra
    store.

    The hydra store is used to construct structured configs from the yaml files.

    We'll also do some preprocessing of the dataclass fields such that all type hints
    are supported by hydra. Hydra only supports a certain subset of types, so we'll
    convert the types to supported types using the _sanitized_type method from
    hydra_zen.

    Keyword Args:
        kw: The kwargs to pass to the dataclass decorator. The following defaults
            are set:

            - repr: False
            - eq: False
            - slots: True
            - kw_only: True
    """

    # Update the kwargs for the dataclass with some defaults
    default_dataclass_kw = dict(repr=False, eq=False, slots=True, kw_only=True)
    kwargs = {**default_dataclass_kw, **kwargs}

    def wrapper(original_cls):
        # Preprocess the fields to convert the types to supported types
        # Only certain primitives are supported by hydra/OmegaConf, so we'll convert
        # these types to supported types using the _sanitized_type method from hydra_zen
        # We'll just include the fields that are defined in this class and not in a base
        # class.
        cls = dataclass(original_cls, **kwargs)

        new_fields = []
        for f in fields(cls):
            # Ignore non-initialized fields
            new_fields.append((f.name, zen.DefaultBuilds._sanitized_type(f.type), f))

        # Create the new dataclass with the sanitized types
        kwargs["bases"] = cls.__bases__
        hydrated_cls = make_dataclass(cls.__name__, new_fields, **kwargs)

        # Copy over custom methods from the original class
        for attr_name in dir(cls):
            if attr_name not in hydrated_cls.__dict__:
                attr = getattr(cls, attr_name)
                if callable(attr):  # Ensure it's a method
                    try:
                        if hasattr(attr, "__func__"):
                            # If it's a bound method, we need to rebind it to the new
                            # class
                            setattr(
                                hydrated_cls,
                                attr_name,
                                types.MethodType(attr.__func__, hydrated_cls),
                            )
                        else:
                            setattr(hydrated_cls, attr_name, attr)
                    except TypeError:
                        pass

        # This is a fix for a bug in the underlying cloudpickle library which is used
        # by hydra/submitit (a hydra plugin) to pickle the configs. Since we're using
        # dataclasses, when pickled, their state doesn't propagate correctly to the new
        # process when it's unpickled. A fix is to define the dataclasses in separate
        # modules, but since we're using make_dataclass all in the same one, we have to
        # explicitly set the module of the class here.
        # See https://github.com/cloudpipe/cloudpickle/issues/386 for a related bug.
        # TODO submit bug report on cloudpickle. #386 is fixed, but _MISSED_TYPE is
        # still an issue.
        hydrated_cls.__module__ = cls.__module__

        # Add to the hydra store
        ConfigStore.instance().store(cls.__name__, hydrated_cls)

        return hydrated_cls

    if cls is None:
        return wrapper
    return wrapper(cls)


# =============================================================================


@config_wrapper
class HydraContainerConfig:
    """Base dataclass which provides additional methods for working with configs."""

    config: Optional[DictConfig] = field(default=None)
    """The original, uninstantiated config. This is maintained within each nested
    instantiated config to allow for proper serialization and deserialization, as well
    as printing the config as a yaml string."""
    custom: Optional[Dict[str, Any]] = field(default_factory=dict)
    """ Custom data to use. This is useful for code-specific logic (i.e. not in yaml
    files) where you want to store data that is not necessarily defined in the config.
    """

    @classmethod
    def instantiate(
        cls,
        config: DictConfig | ListConfig,
        *,
        _convert_: str = "object",
        **kwargs,
    ) -> Self:
        """Instantiate the config into an object.

        Args:
            config (DictConfig | ListConfig): The config to instantiate.

        Keyword Args:
            _convert_ (str): The conversion method to use. Defaults to "object", meaning
                all structured configs will be converted to their dataclass equivalent.
            **kwargs: Additional keyword arguments to pass to the instantiation method.
        """

        instance: Self = zen.instantiate(config, _convert_=_convert_, **kwargs)
        OmegaConf.resolve(config)

        # Iteratively set the config attribute for all nested configs
        def set_config_attr(obj: Any, config: DictConfig | ListConfig):
            if isinstance(obj, HydraContainerConfig):
                if obj.config is None:
                    obj.config = config
                if not OmegaConf.is_config(obj.config):
                    obj.config = OmegaConf.create(obj.config)
                for k, v in obj.config.items():
                    if hasattr(obj, k):
                        set_config_attr(getattr(obj, k), v)
            elif isinstance(obj, dict):
                for k, v in obj.items():
                    if k in config:
                        set_config_attr(v, config[k])
            elif isinstance(obj, list):
                for i, v in enumerate(obj):
                    set_config_attr(v, config[i])

        # After instantiation, we'll set the config attribute for all nested configs
        # `config` is ignored by omegaconf, so has to come after initialization
        set_config_attr(instance, config)

        return instance

    @classmethod
    def compose(
        cls,
        config_dir: Path | str,
        config_name: str,
        *,
        overrides: List[str] = [],
        return_hydra_config: bool = False,
        **kwargs,
    ) -> (
        Self
        | DictConfig
        | ListConfig
        | Tuple[Self | DictConfig | ListConfig, DictConfig]
    ):
        """Compose a config using the Hydra compose API. This will return the config as
        a HydraContainerConfig instance.

        Args:
            config_dir (Path | str): The path to the config directory.
            config_name (str): The name of the config file.

        Keyword Args:
            overrides (List[str]): The overrides to use when composing the config.
            return_hydra_config (bool): Whether to return the HydraConfig object.
            **kwargs: Additional keyword arguments to pass to the instantiation method.
        """

        import hydra
        from hydra.core.global_hydra import GlobalHydra
        from hydra.core.hydra_config import HydraConfig

        if GlobalHydra.instance().is_initialized():
            GlobalHydra.instance().clear()

        with hydra.initialize_config_dir(str(config_dir), version_base=None):
            hydra_config = hydra.compose(
                config_name=config_name, overrides=overrides, return_hydra_config=True
            )
            HydraConfig.instance().set_config(hydra_config)
            del hydra_config.hydra

        config = cls.create(hydra_config, **kwargs)
        if return_hydra_config:
            return config, HydraConfig.get()
        return config

    @classmethod
    def load(
        cls,
        *args,
        instantiate: bool = True,
        pattern: Optional[str] = None,
        **instantiate_kwargs,
    ) -> Self | DictConfig | ListConfig:
        """Wrapper around OmegaConf.load to instantiate the config.

        Keyword Args:
            instantiate (bool): Whether to instantiate the config into an object.
            pattern (Optional[str]): The specific pattern to select from the loaded
                config.
            **instantiate_kwargs: Additional keyword arguments to pass to the
                instantiation method.
        """
        loaded = OmegaConf.load(*args)
        if pattern is not None:
            loaded = OmegaConf.select(loaded, pattern)
        if instantiate:
            return cls.instantiate(loaded, **instantiate_kwargs)
        else:
            return loaded

    @classmethod
    def create(
        cls, *args, instantiate: bool = True, **instantiate_kwargs
    ) -> Self | DictConfig | ListConfig:
        """Wrapper around OmegaConf.create to instantiate the config.

        Keyword Args:
            instantiate (bool): Whether to instantiate the config into an object.
            **instantiate_kwargs: Additional keyword arguments to pass to the
                instantiation method.
        """
        created = OmegaConf.create(*args)
        if instantiate:
            return cls.instantiate(created, **instantiate_kwargs)
        else:
            return created

    def merge_with(self, *others: DictConfig | ListConfig | Dict | List) -> Self:
        """Wrapper around OmegaConf.merge to merge the config with another config.

        Args:
            others (DictConfig | ListConfig | Dict | List): The other config(s) to
                merge with.
        """
        # Do an unsafe merge so types aren't checked
        merged = OmegaConf.unsafe_merge(self.config, *others)
        return self.instantiate(merged)

    def copy(self) -> Self:
        """Wrapper around the copy method to return a new instance of this class.

        Note:

            This method will perform a deepcopy, meaning the :meth:`__getstate__` and
            :meth:`__setstate__` methods will be called. This is fairly slow since
            the object is pickled and unpickled.
        """
        return deepcopy(self)

    def save(
        self,
        path: Path | str,
        *,
        header: str = None,
    ):
        """Saves the config to a yaml file.

        Args:
            path (Path | str): The path to save the config to.

        Keyword Args:
            header (str): The header to add to the top of the yaml file.
        """

        with open(path, "w") as f:
            if header:
                f.write(f"{header}\n")
            f.write(self.to_yaml())

    def to_yaml(self) -> str:
        """Wrapper around OmegaConf.to_yaml to convert the config to a yaml string.
        Adds some custom representers.

        This uses the stored config attribute to convert to yaml. If the config is None,
        this will return the default string representation of the object.
        """

        assert self.config is not None, "Config is None, cannot convert to yaml."

        def str_representer(dumper: yaml.Dumper, data: str):
            style = None
            if "\n" in data:
                # Will use the | style for multiline strings.
                style = "|"
            elif data == MISSING:
                # Will wrap ??? in quotes, yaml doesn't like this otherwise
                style = '"'
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=style)

        def path_representer(dumper: yaml.Dumper, data: Path):
            return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))

        def flag_representer(dumper: yaml.Dumper, data: enum.Flag):
            data = "|".join([m.name for m in type(data) if m in data])
            return dumper.represent_scalar("tag:yaml.org,2002:str", data)

        dumper = yaml.CDumper
        dumper.add_representer(str, str_representer)
        dumper.add_multi_representer(Path, path_representer)
        dumper.add_multi_representer(enum.Flag, flag_representer)
        config = OmegaConf.to_container(self.config)
        return yaml.dump(
            config,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
            Dumper=dumper,
        )

    def __getstate__(self) -> DictConfig:
        """This is used to pickle the object. We'll return the config as the state."""
        return self.config

    def __setstate__(self, state: DictConfig):
        """This is used to unpickle the object. We'll set the config from the state."""
        instance = self.instantiate(state)
        for field_name in self.__dataclass_fields__.keys():
            setattr(self, field_name, getattr(instance, field_name))

    def __str__(self) -> str:
        """Convert the config to a yaml string."""
        if self.config is None:
            return self.__repr__()
        return self.to_yaml()
